{  _  )(_  _)(_  _)( ___) Object-oriented
 )(_)(   )(    )(   )__)  Tengwar
(_____) (__)  (__) (____) Transcription
 VERSION 0.0.5            Engine

a GPL software by <miszka@limes.com.pl>

OTTE is a library that provides you with easy
way to create programs like TengScribe or YaTT
in Delphi or FPC. Although it's not as functional
as YaTT (no transcription back to latin), it's
even faster, and still in development, so there
may be new functionalities added. For now, see
the program TTU to learn how to use OTTE.
-----------------------------------------------}
{CONFIGURATION}
{$IFDEF VER100}{$DEFINE DELPHI}{$ENDIF}
{$IFDEF VER110}{$DEFINE DELPHI}{$ENDIF}
{$IFDEF VER120}{$DEFINE DELPHI}{$ENDIF}
{$IFDEF VER130}{$DEFINE DELPHI}{$ENDIF}
{$IFDEF VER140}{$DEFINE DELPHI}{$ENDIF}
{$IFDEF VER150}{$DEFINE DELPHI}{$ENDIF}

{DEFINE DEBUG}

{$IFDEF DELPHI}
  {$DEFINE WINDOWS}
  {$I-,B-,J+,P-,V-,H-}

  {$IFDEF DEBUG}{$D+,L+,Q+,Y+,R+,S+}
  {$ELSE}{$D-,L-,Q-,Y-,R-,S-}
  {$ENDIF}
{$ENDIF}

{$IFDEF FPC}
  {$SMARTLINK ON}
  {$MODE DELPHI}
  {$IFNDEF LINUX}{$S-}{$ENDIF}
  {$I-,P-,V-,M-,H+,B-,X+}

  {$IFDEF DEBUG}{$D+,L+,Y+,Q+,R+}
  {$ELSE}{$D-,L-,Y-,Q-,R-}
  {$ENDIF}
{$ENDIF}
//----------------------------------------------

unit otte;

interface

uses
  sysalpha;

const
  CtteErrLocInitFailed= -1;
  CtteErrModNotOpened = -2;
  CtteErrModReadFault = -3;
  CtteErrModTooLarge  = -4;
  CtteErrModBadHeader = -5;
  CtteErrModBadFormat = -6;
  CtteErrNotEnoughRAM = -7;
  CtteErrModNotLoaded = -8;
  
  CtteLocAsRoman = true;
  CtteLocAsTengwar = false;

  CtteLocDefaultQuestionMark = #174;

  CtteModStringLen = 14;
  CtteModClauseSize = 36;
  CtteModDescLen = CtteModClauseSize-2;
  CtteModMaxClauses = 1500;

  CtteModAllChar = #00;
  CtteModNoAlpha = #13;

type
  TtteLocCharSet = array[char] of boolean;
  TtteLocIsNotAlpha = array[boolean] of TtteLocCharSet;
  TtteLocToLower = array[char] of char;

  TtteLoc = class(TObject)
    public
      IsNotAlpha: TtteLocIsNotAlpha;
      ToLower: TtteLocToLower;
      QuestionMark: char;
      procedure reset;
      constructor create;
      function StrLower(const s: ansistring): ansistring;
      function load(const locname: ansistring): integer;
  end;

  TtteModDesc = string[CtteModDescLen];
  TtteModString = string[CtteModStringLen];

  TtteModClause = record
    RomanMod, TengwarMod: TtteModString;
    NextRoman, LastTengwa, SetLastTengwa: char;
  end;

  TtteModClauses = array of TtteModClause;
  TtteModSortControl = function(const a, b: TtteModClause): boolean;

  TtteMode = class(TObject)
    public
      locale: TtteLoc;         //please first of all load locale, then mode
      clauses: TtteModClauses;
      description: TtteModDesc;
      CaseSensitive: boolean;
      procedure reset;
      constructor create;
      destructor destroy; override;
      procedure sort(const control: TtteModSortControl);
      function load(const fname: ansistring): integer;
  end;

  TtteTransEngine = class(TObject)
    protected
      FirstInd,
      LastInd: array[char] of integer;
      IsWhiteSpace: array[char] of boolean;
    public
      RevNum: boolean;        //reverse digit order
      CirNum: boolean;        //mark duedecades with circles beneath
      DotNum: boolean;        //mark numerals with dots
      BarNum: boolean;        //mark numerals with bars
      B12Num: boolean;        //use duodecimal numeric base
      TxtAll: boolean;        //don't treat numbers or spaces separately
      mode: TtteMode;
      procedure reset;
      constructor create;
      destructor destroy; override;
      function GetReady: integer;  //run this before any transcription task
      function TranscribeLine(const s: ansistring): ansistring;
  end;

const
  CtteLocDefaultIsNotTengwa: TtteLocCharSet =
  ( true,  true,  true,  true,  true,  true,  true,  true,
    true,  true,  true,  true,  true,  true,  true,  true,
    true,  true,  true,  true,  true,  true,  true,  true,
    true,  true,  true,  true,  true,  true,  true,  true,
    true,  false, false, false, false, false, false, false,
    false, false, false, false, false, true,  false, false,
    false, false, false, false, false, false, false, false,
    false, false, false, false, false, true,  false, false,
    false, false, false, false, false, false, false, false,
    false, false, false, false, false, false, false, false,
    false, false, false, false, false, false, false, false,
    false, false, false, false, false, false, false, false,
    false, false, false, false, false, false, false, false,
    false, false, false, false, false, false, false, false,
    false, false, false, false, false, false, false, false,
    false, false, false, false, false, false, false, true,
    true,  true,  false, false, false, false, false, false,
    true,  false, false, false, true,  true,  true,  true,
    true,  false, false, false, false, false, false, false,
    false, false, false, true,  true,  true,  true,  false,
    true,  false, false, false, true,  false, false, false,
    false, false, false, true,  true,  false, true,  false,
    false, true,  true,  false, false, false, true,  true,
    false, true,  false, true,  false, false, false, false,
    true,  true,  true,  true,  false, false, false, true,
    false, false, false, false, false, false, false, false,
    false, false, false, false, false, false, false, false,
    false, false, false, false, false, false, false, false,
    false, false, false, false, false, false, false, false,
    false, false, false, false, false, false, false, false,
    true,  true,  true,  true,  true,  true,  true,  true,
    true,  true,  true,  true,  false, false, false, false  );

implementation
//===TtteLoc===================================================================
procedure TtteLoc.reset;
begin
  self.IsNotAlpha[CtteLocAsTengwar] := CtteLocDefaultIsNotTengwa;
  self.QuestionMark := CtteLocDefaultQuestionMark;
  self.load('');
end;

//-----------------------------------------------------------------------------
constructor TtteLoc.create;
begin
  inherited;
  self.reset;
end;

//-----------------------------------------------------------------------------
function TtteLoc.StrLower(const s: ansistring): ansistring;

var
  i: integer;

begin
  result := s;
  for i := 1 to length(s) do result[i] := ToLower[s[i]];
end;

//-----------------------------------------------------------------------------
function TtteLoc.load(const locname: ansistring): integer;

var
  c: char;

begin
  if sysalpha.setlocale(locname) then begin
    result := 0;
    
    for c := #0 to #255 do begin
      self.IsNotAlpha[CtteLocAsRoman, c] := not sysalpha.isalpha(c);
      self.ToLower[c] := sysalpha.tolower(c);
    end;
  end
  else result := CtteErrLocInitFailed;
end;

//===TtteMode==================================================================
procedure TtteMode.reset;
begin
  locale.reset;
  SetLength(clauses, 0);
  description := '';
  CaseSensitive := false;
end;

//-----------------------------------------------------------------------------
constructor TtteMode.create;
begin
  inherited;
  locale := TtteLoc.create;
  reset;
end;

//-----------------------------------------------------------------------------
destructor TtteMode.destroy;
begin
  reset;
  locale.destroy;
  inherited;
end;

//-----------------------------------------------------------------------------
procedure TtteMode.sort(const control: TtteModSortControl);

var
  i: integer;
  changed: boolean;
  c: TtteModClause;

{Don't use any 'unstable' sorting method!}
begin
  if Addr(control) = nil then exit;

  repeat
    changed := false;

    for i := 0 to high(clauses)-1 do
    if control(clauses[i], clauses[i+1]) then begin
      changed := true;
      c := clauses[i];
      clauses[i] := clauses[i+1];
      clauses[i+1] := c;
    end;
  until not changed;
end;

//-----------------------------------------------------------------------------
function TtteMode.load(const fname: ansistring): integer;

type
  TtteDecodeRecData = array[0..CtteModClauseSize-1] of byte;

  //. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .
  function chop(var s: ansistring): ansistring;
  begin
    result := '';

    while (s<>'') and (s[1]<>#09) do begin
      result := result + s[1];
      delete(s, 1, 1);
    end;

    if s<>'' then delete(s, 1, 1);
  end;
  //. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .
  function DecodeHeader(
    const data: TtteDecodeRecData;
    var Description: TtteModDesc;
    var CaseSensitive: boolean
  ): integer;

  var
    i: integer;
    s, desc, csflag: ansistring;

  begin
    result := 0;

    s := '';
    for i := 0 to CtteModClauseSize-1 do s := s + char(data[i]);

    desc := chop(s);
    if length(desc) > CtteModDescLen then begin
      result := CtteErrModBadFormat;
      exit;
    end;

    csflag := chop(s);
    if (csflag<>'0') and (csflag<>'1') then begin
      result := CtteErrModBadFormat;
      exit;
    end;

    Description := desc;
    if csflag = '0' then CaseSensitive := false
    else if csflag = '1' then CaseSensitive := true
  end;
  //. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .
  function DecodeClause(
    const data: TtteDecodeRecData;
    var clause:  TtteModClause
  ): integer;

  var
    i: integer;
    s, rommod, tengmod, nextr, lastt, setlastt: ansistring;

  begin
    result := 0;

    s := '';
    for i := 0 to CtteModClauseSize-1 do s := s + char(data[i]);    

    rommod := chop(s);
    if (rommod='') or (length(rommod)>CtteModStringLen) then begin
      result := CtteErrModBadFormat;
      exit;
    end;
   
    tengmod := chop(s);
    if (length(tengmod)>CtteModStringLen) then begin
      result := CtteErrModBadFormat;
      exit;
    end;
    
    nextr := chop(s);
    if (nextr='') then nextr := CtteModAllChar;

    lastt := chop(s);
    if (lastt='') then lastt := CtteModAllChar;
    
    setlastt := chop(s);
    if (setlastt='') then setlastt := CtteModAllChar;

    with clause do begin
      RomanMod := rommod;
      TengwarMod := tengmod;
      NextRoman := nextr[1];
      LastTengwa := lastt[1];
      SetLastTengwa := setlastt[1];
      
      if self.CaseSensitive then begin
        RomanMod := self.locale.StrLower(RomanMod);
	NextRoman := self.locale.ToLower[NextRoman];
      end;
    end;
  end;
  //. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

var
  f: file;
  fsize, count: integer;
  data: array of TtteDecodeRecData;
  TempClauses: TtteModClauses;
  ClausesCount: integer;
  i: integer;
  desc: TtteModDesc;
  csflag: boolean;

begin
  result := 0;
  
  assign(f, fname);
  system.reset(f, 1);

  if IOResult<>0 then begin
    result := CtteErrModNotOpened;
    exit;
  end;
  
  fsize := FileSize(f);
  ClausesCount := fsize div CtteModClauseSize - 2;

  if ((fsize mod CtteModClauseSize)<>0) or (ClausesCount<=0) then 
    result := CtteErrModBadFormat
  else if ClausesCount > CtteModMaxClauses then
    result := CtteErrModTooLarge;

  if result<>0 then begin
    close(f);
    exit;
  end;

  try
    SetLength(data, ClausesCount+2);
    SetLength(TempClauses, ClausesCount);
  except
    {
      Code below gets repeated and repeated.
      I must rewrite it once.
    }
    result := CtteErrNotEnoughRAM;
    close(f);
    SetLength(data, 0);
    SetLength(TempClauses, 0);
    exit;
  end;

  BlockRead(f, data[0], fsize, count);

  if (IOResult<>0) or (count<>fsize) then begin
    result := CtteErrModReadFault;
    close(f);
    SetLength(data, 0);
    SetLength(TempClauses, 0);
    exit;
  end;
  
  close(f);
  
  result := DecodeHeader(data[0], desc, csflag);
  
  if result<>0 then begin
    SetLength(data, 0);
    SetLength(TempClauses, 0);
    exit;
  end;
   
  for i := 2 to ClausesCount+1 do begin
    result := DecodeClause(data[i], TempClauses[i-2]);
    
    if result<>0 then begin
      SetLength(data, 0);
      SetLength(TempClauses, 0);
      exit;
    end;
  end;

  SetLength(data, 0);
 
  if result=0 then begin
    self.CaseSensitive := csflag;
    self.description := desc;
    self.clauses := TempClauses;
    {
    Don't finalize TempClauses!
    That will erase our new self.clauses!
    For similar, don't finalize old self.clauses!
    }
  end
  else SetLength(TempClauses, 0);
end;

//===TtteTransEngine===========================================================
procedure TtteTransEngine.reset;
begin
  mode.reset;
  RevNum := false;
  CirNum := false;
  DotNum := false;
  BarNum := false;
  B12Num := false;
  TxtAll := false;
end;

//-----------------------------------------------------------------------------
constructor TtteTransEngine.create;
begin
  inherited;
  mode := TtteMode.create;
  reset;
end;

//-----------------------------------------------------------------------------
destructor TtteTransEngine.destroy;
begin
  mode.destroy;
  inherited;
end;

//-----------------------------------------------------------------------------
function SortControl(const a, b: TtteModClause): boolean; register;
begin
  result := a.RomanMod[1] > b.RomanMod[1];
end;

//-----------------------------------------------------------------------------
function TtteTransEngine.GetReady: integer;
{
  Run this proc whenever you have just set all the params needed to do
  the transcription and before you start. If you don't, you'll get
  erroneous results!

  This proc prepares some data for transcription process. It also sorts
  and indexes the mode clauses by their first letters to speed the
  transcription process up (about 1.5 times).
}

var
  c: char;
  i: integer;
  cs: TtteModSortControl;

begin
  if length(mode.clauses) = 0 then begin
    result := CtteErrModNotLoaded;
    exit;
  end
  else result := 0;

  //Set characters skipped in transcription process
  fillchar(IsWhiteSpace, SizeOf(IsWhiteSpace), false);
  fillchar(IsWhiteSpace, 32, true);

  if not TxtAll then begin
    IsWhiteSpace[#32] := true;
    IsWhiteSpace[#160] := true;
  end;

  //Sort the mode clauses
  cs := SortControl;
  mode.sort(cs);

  //Index the mode clauses
  //chars that do not open any mode rule are given strange
  //indexes to be immediately detected during thanscription
  for c := #00 to #255 do begin
    FirstInd[c] := high(mode.clauses);
    
    for i := 0 to high(mode.clauses) do begin
      if mode.clauses[i].RomanMod[1] = c then begin
        FirstInd[c] := i;
	break;
      end;
    end;

    LastInd[c] := 0;
    
    for i := high(mode.clauses) downto 0 do begin
      if mode.clauses[i].RomanMod[1] = c then begin
        LastInd[c] := i;
	break;
      end;
    end;
  end;
end;

//-----------------------------------------------------------------------------
function TtteTransEngine.TranscribeLine(const s: ansistring): ansistring;

var
  buffer: ansistring;
  LastWrittenTengwa: char;
  index: integer;
  {$IFDEF WINDOWS}ousindex: integer;{$ENDIF}

  //. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .
  {$IFDEF WINDOWS}
  procedure StrAdd(const ins: ansistring; var ousindex: Integer; var ous: ansistring);

  var
    count: integer;

  begin
    count := length(ins);
    if (ousindex+count > length(ous)) then SetLength(ous, ousindex+count+80);
    move(ins[1], ous[ousindex], count);
    inc(ousindex, count);
  end;
  {$ENDIF}
  //. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .
  function IsDigit(c: char): boolean;
  begin
    result := c in ['0'..'9'];
  end;
  //. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .
  function ProcessNumerals(const buffer: ansistring; var index: integer): ansistring;

  const
    MaxB12Len = 18; //int64 will deal

    //indexes  for Nums[digit, index] array
    //Note for barover and barunder: they are taken in account only for first
    //written digit, as the bar  may be longer or shorter. Then, for the second
    //and following digits, the standard, long bars are used.
    numeral  = 0;
    dotover  = 1;
    barover  = 2;
    dotunder = 3;
    barunder = 4;
    circle   = 5;

    // standard bars
    LongBarOver  = #236;  //over second and further digits
    LongBarUnder = #237;  //under second and further digits

    // Array of Elvish numerals and their "attributes"
    Nums: array[0..11, numeral..circle] of char =

     // numeral dotover barover dotunder barunder circle      digit
        ((#240,   'T',   #238,   #201,    #239,   #168),     {  0  }
         (#241,   'G',   #238,   #202,    #239,   #168),     {  1  }
         (#242,   '%',   #236,   #200,    #237,   #153),     {  2  }
         (#243,   '%',   #236,   #200,    #237,   #152),     {  3  }
         (#244,   'G',   #238,   #202,    #239,   #153),     {  4  }
         (#245,   'T',   #238,   #201,    #239,   #152),     {  5  }
         (#246,   '%',   #236,   #200,    #237,   #152),     {  6  }
         (#247,   'T',   #238,   #201,    #239,   #153),     {  7  }
         (#248,   'G',   #238,   #202,    #239,   #153),     {  8  }
         (#249,   '%',   #236,   #200,    #237,   #152),     {  9  }
         (#250,   'B',   #238,   #202,    #239,   #153),     {  10 }
         (#251,   'B',   #238,   #202,    #239,   #153));    {  11 }


    //array of digits. Note, that for digits '10' and '11' I used ':' and ';',
    //because they are after '9' in the ASCII, aqnd that makes conversion to
    //duodecimal format easier.
    Digits: array[0..11] of char =
            ( '0', '1', '2', '3', '4', '5', '6', '7', '8', '9', ':', ';' );
  var
    NumBuff: ansistring;    //numeric ASCII string
    NumVal: Int64;          //it holds the numeral value when processed to duodecimal
    count: integer;         //length of data read
    i: integer;
    k: integer;             //we write NumBuff[k], not NumBuff[i]
    NumLen: Integer;        //length of input numeral
    j: Byte;                //NumBuff[k] converted into index of Nums array

  begin
    NumBuff := '';
    Result := '';

    //read numer al from buffer into
    count := 0;
    while (index+count <= length(buffer)) and IsDigit(buffer[index+count]) do inc(count);
    NumBuff := Copy(buffer, index, count);
    inc(index, count);

    //convert into duodecimal, if requested.
    //"0" is the trivial case that don't need to be converted
    if B12Num and (NumBuff<>'0') then begin
      if length(NumBuff) > MaxB12Len then SetLength(NumBuff, MaxB12Len);
  
      Val(NumBuff, NumVal, i);
      NumBuff := '';

      while NumVal <> 0 do begin
        Insert(Digits[NumVal mod 12], NumBuff, 1);
        NumVal := NumVal div 12;
      end;
    end;

    NumLen := Length(NumBuff);

    //write the numeral now
    for i := 1 to NumLen do begin
      if RevNum then k := NumLen-i+1 else k := i; //do U want reverse order?

      //write the digit
      j := Byte( NumBuff[k] ) - Ord('0');
      Result := Result + Nums[j, numeral];

      //if dots requested
      if DotNum then begin
        if B12Num then begin
          if (not CirNum) or (k <> NumLen-1) then
            Result := Result + Nums[j, dotunder];
        end
        else Result := Result + Nums[j, dotover];
      end;

      //if bars requested
      if BarNum then begin
        if (k=1) then begin //first written digit - use sophisticated bars
          if B12Num then Result := Result + Nums[j, barunder]
          else Result := Result + Nums[j, barover];
        end
        else begin //use standard long bars
          if B12Num then Result := Result + LongBarUnder
          else Result := Result + LongBarOver;
        end;
      end;

      //if  circled duodecades requested
      if CirNum and (k = NumLen-1) then Result := Result + (Nums[j, circle]);
    end;
  end;
  //. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .
  function CharsMatch(ch, ch_mod: char; AsRoman: boolean): boolean;
  begin
    Result := (ch_mod = CtteModAllChar) or (ch = ch_mod) or
              ((ch_mod = CtteModNoAlpha) and mode.locale.IsNotAlpha[AsRoman, ch]);
  end;
  //. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .
  function ProcessText(const buffer: ansistring; var index: integer): TtteModString;

  var
    i: Integer;
    Temp: String[CtteModStringLen+1];
    modclause: ^TtteModClause;

  begin
    Result := '';
    Temp := '               ';
    Temp := copy(buffer, index, CtteModStringLen+2);
    modclause := addr(mode.clauses[FirstInd[Temp[1]]-1]);

    //here GetReady indexing efforts take effect
    //we don't have to search through all the mode clauses
    for i := FirstInd[Temp[1]] to LastInd[Temp[1]] do begin
      Inc(modclause);

      With modclause^ do begin
        Temp[0] := RomanMod[0];

        //further speed improvement - great for complicated modes (i.e. Quenya)
	//caution - checking characters no. 4 and next slows things down...
	//takes no effect for simple modes (i. e. Sindarin_beleriand)
        if (Temp[0] >= #2) and (RomanMod[2] <> Temp[2]) then continue
        else if (Temp[0] >= #3) and (RomanMod[3] <> Temp[3]) then continue
        else if
           CharsMatch(LastWrittenTengwa, LastTengwa, CtteLocAsTengwar) and
           CharsMatch(Temp[Byte(Temp[0])+1], NextRoman, CtteLocAsRoman) and
	   (RomanMod = Temp) then
        begin
          result := TengwarMod;
          LastWrittenTengwa := SetLastTengwa;
	  inc(index, length(RomanMod));
          exit;
        end;
      end;
    end;

    // when transcription error...
    result := mode.locale.QuestionMark;
    LastWrittenTengwa := ' ';
    inc(index);
  end;
  //. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

begin
  result := '';
  if length(mode.clauses) = 0 then exit;
  index := 1;
  
  {$IFDEF WINDOWS}
  ousindex := 1;
  setLength(result, length(s));
  {$ENDIF}
  
  LastWrittenTengwa := ' ';
  
  if mode.CaseSensitive then buffer := s
  else buffer := mode.locale.StrLower(s);

  while index <= length(buffer) do  begin
    if IsWhiteSpace[buffer[index]] then begin
      {$IFDEF WINDOWS}StrAdd(buffer[index], ousindex, result);
      {$ELSE}result := result + buffer[index];
      {$ENDIF}
      inc(index);
      LastWrittenTengwa := ' ';
    end
    else if (not TxtAll) and IsDigit(buffer[index]) then begin
      {$IFDEF WINDOWS}StrAdd(ProcessNumerals(buffer, index), ousindex, result);
      {$ELSE}result := result + ProcessNumerals(buffer, index);
      {$ENDIF}
      LastWrittenTengwa := CtteModNoAlpha;
    end
    else begin
      {$IFDEF WINDOWS}StrAdd(ProcessText(buffer, index), ousindex, result);
      {$ELSE}result := result + ProcessText(buffer, index);
      {$ENDIF}
    end;
  end;
  
  {$IFDEF WINDOWS}SetLength(result, ousindex-1);{$ENDIF}
end;

//===EndOfImplementation=======================================================
end.
